package polyglot.ext.polyj.parse;

import polyglot.ast.*;
import polyglot.visit.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.parse.*;

import polyglot.ext.polyj.ast.*;
import polyglot.ext.polyj.types.*;

import java.util.*;

/**
 * A <code>Parse</code> is a piece of syntax for which the AST representation
 * is considered ambiguous until more is known about its context.   
 * 
 * These pieces are:
 * P ::= n | P.n | P[] | P[L]
 * L :: = T | T, L
 * where P is a Parse, n is an identifier, and T is a type or expression 
 * represented by a subclass of Sub.
 * 
 * Subclasses of <code>Parse</code> encapsulate the data necessary to convert 
 * the syntax into the proper AST node, as well as the methods to perform the conversion.
 * <code>Parse</code> itself implements these methods by causing a fatal parsing error.
 * The effect is that if a subclass doesn't override a conversion method, the conversion
 * is illegal on that subtype.
 * 
 * No Parse node should ever make its way into the AST, unless wrapped via the Wrapper class.
 */
public abstract class Parse {
    Position pos;
    BaseParser parser;
    PolyJNodeFactory nf;
    PolyJTypeSystem ts;

    public Parse(BaseParser parser, Position pos) {
        this.parser = parser;
        this.pos = pos;
        this.nf = (PolyJNodeFactory) parser.nf;
        this.ts = (PolyJTypeSystem) parser.ts;
    }

    /**
     * Wrap this to make it a legal AST node.
     */
    public Expr wrap() throws Exception { 
        return new Wrapper(this); 
    }
    
    /**
     * Convert this into a Call, or cause the parser to die
     * if the conversion is illegal.
     */
    public Call toCall(List args) throws Exception {
        parser.die("Illegal method call.", pos); 
        return null;
    }
    
    /**
     * Convert this into an Expr.  The precise subclass
     * of Expr will be determined by the overriding method.
     * Cause the parser to die if the conversion is illegal.
     */
    public Expr toExpr() throws Exception {
        parser.die("Illegal expression.", pos); 
        return null; 
    }
    
    /**
     * Convert this into a PackageNode, or cause the parser to die
     * if the conversion is illegal.
     */
    public PackageNode toPackage() throws Exception { 
        parser.die("Illegal package.", pos); 
        return null; 
    }
    
    /**
     * Convert this into a Prefix, which is an expression, type, or package, and
     * prefixes a Receiver.
     */
    public Prefix toPrefix() throws Exception { 
        parser.die("Illegal prefix.", pos); 
        return null; 
    }
    
    /**
     * Convert this into a Receiver, which is an expression or a type, and prefixes
     * an expression.
     */
    public Receiver toReceiver() throws Exception { 
        parser.die("Illegal receiver.", pos); 
        return null; 
    }
    
    /**
     * Convert this into a QualifierNode, which is a type or a package, and prefixes
     * a type.
     */
    public QualifierNode toQualifier() throws Exception {
        parser.die("Illegal qualifier.", pos);
        return null;
    }
    
    /**
     * Convert this into a TypeNode, or cause the parser to die
     * if the conversion is illegal.
     */
    public TypeNode toType() throws Exception { 
        parser.die("Illegal type.", pos); 
        return null; 
    }
    
    /**
     * Convert this into a TypeNode that contains a ClassType.
     */
    public TypeNode toClassType() throws Exception { 
        parser.die("Illegal class type.", pos); 
        return null; 
    }
    
    /**
     * Convert this into a new array expression (e.g., new int[10]).
     */
    public NewArray toNewArray(Position p, int addDims, ArrayInit init) throws Exception { 
        parser.die("Illegal new array expression.", p); 
        return null; 
    }
    
    /**
     * Convert this into a new array expression (e.g., new int[10]).
     */
    public NewArray toNewArray(Position p, int addDims) throws Exception { 
        return toNewArray(p, addDims, null);
    }
    
    /**
     * Convert this into a new array expression (e.g., new int[10]).
     */
    public NewArray toNewArray(Position p) throws Exception { 
        return toNewArray(p, 0, null);
    }
    
    /**
     * Convert this into the prefix of a new array expression 
     * (e.g., someClass.new innerClass[10]).
     */
    public Expr toNewArrayPrefix(Position p) throws Exception { 
        parser.die("Illegal prefix for new array expression.", pos); 
        return null; 
    }
    
    /**
     * Convert this into an identifier, which is a single Java 
     * identifier.
     */
    public String toIdentifier() throws Exception { 
        parser.die("Illegal identifier.", pos); 
        return null; 
    }
    
    /**
     * Convert this into a name, which is a series of identifiers
     * separated by dots (a qualified name).
     */
    public String toName() throws Exception { 
        parser.die("Illegal name.", pos); 
        return null; 
    }
    
    /**
     * Build a <code>Parse</code> out of a subscript syntax -- this[b] -- creating the most
     * specific type possible.  
     * @param subs A list of Subs.
     */
    public Parse append(List subs) throws Exception {
        if (subs.size() == 1) {
            Sub s = (Sub) subs.get(0);
            return s.prepend(this);
        } else {
            return new Inst(parser, pos, toClassType(), subs);
        }
    }

    /**
     * Prepend another Parse to this, creating base[this].
     */
    public Parse prepend(Parse base) throws Exception {
        parser.die("Illegal instantiation or access."
                       + " (prepend on a " + base.getClass().getName() + " )", pos);
        return null;
    }
    
    /**
     * Wrap this in an AmbSub.
     */
    public AmbSub toAmbSub() {
        return new AmbSub(parser, pos, this);
    }
    
    /**
     * Convert this to a parameter.
     */
    public Receiver toParam() throws Exception {
        parser.die("Illegal parameter.", pos); 
        return null;         
    }
    
    /**
     * Convert this to a ParamNode
     */
    public ParamNode toParamNode() throws Exception {
        return nf.Param(pos, toParam());
    }
     
}

