package polyglot.ext.polyj.types;

import polyglot.types.*;
import polyglot.types.Package;
import polyglot.util.*;

import polyglot.ext.polyj.util.*;
import polyglot.ext.polyj.visit.*;

import java.util.*;

public class PolyJTranslationHelper implements TranslationTable {

    /* Set[ClassType]*/
    protected Set instantiations;

    /* ParamType -> List[Integer] */
    protected Map newArrayMethods;

    /* PolyJClassType -> ClassType -> AdapterBinding*/
    protected Map adapterBindings;

    /* PolyJClassType -> Integer */
    private Map nextAdapterMethodBindingNum;

    protected PolyJTypeSystem ts;

    public PolyJTranslationHelper(PolyJTypeSystem ts) {
        this.ts = ts;
        instantiations = new HashSet();
        newArrayMethods = new HashMap();
        adapterBindings = new HashMap();

        nextAdapterMethodBindingNum = new HashMap();
    }

    /**
     * Lookup the given key in the given map,
     * if key not found, create a new entry with the given key
     * if createIfNotFound is true.
     * Here the range of map should consist of Map objects.
     */
    static Map submapForKey(Map map, Object key,
                            boolean createIfNotFound) {
        Object value = map.get(key);
        if (createIfNotFound && value == null) {
            /* so create a new one... */
            value = new HashMap();
            map.put(key, value);
        };
         return (Map)value;
    }

    static Set setForKey(Map map, Object key, boolean createIfNotFound) {
        Object value = map.get(key);
        if (value == null && createIfNotFound) {
            value = new TreeSet();
            map.put(key, value);
        }
        return (Set) value;
    }

    /**
     * Record that <code>type</code> was instantiated.
     */
    void addInstantiation(ClassType type) {
        instantiations.add(type);
    }

    void addNewArrayMethod(ParamType param, int dims) {
        Set s = setForKey(newArrayMethods, param, true);
        s.add(new Integer(dims));
    }

    void addAssistedInstantiation(PolyJClassType assistant,
        ClassType instantiation)
    {
        Map m = submapForKey(adapterBindings, assistant, true);
        m.put(instantiation, ts.AdapterBinding(assistant,
           instantiation));
    }

    /********************************************************************/
    /* Methods implementing the TranslationTable interface              */
    /********************************************************************/

    public Set instantiations() {
        return instantiations;
    }

    public Iterator newArrayMethods(ParamType param) {
        Set s = setForKey(newArrayMethods, param, false);
        if (s == null) {
            return Collections.EMPTY_SET.iterator();
        } else {
            return s.iterator();
        }
    }

    /**
     * Returns a list of adapter bindings for all the classes that
     * the given polymorphic type instantiates
     */
    public Iterator/*of AdapterBinding*/ getAdapterBindings(ClassType poly) {
        Map m =  submapForKey(adapterBindings, poly, false);
        if (m == null)
            return Collections.EMPTY_SET.iterator();
        else
            return m.values().iterator();
    }

    /**
     * Returns the adapter binding for the given assisted instantiation
     */
    public AdapterBinding getAdapterBinding(PolyJClassType assistant,
                                            ClassType inst) {
        Map m = submapForKey(adapterBindings, assistant, false);
        if (m == null)
            return null;
        else
            return (AdapterBinding) m.get(inst);
    }

    /**
     * Return the class whose where object must be created by
     * <code>inst</code>'s adapter class for <code>adapter</code>.
     *
     * As an example, suppose we have:
     *
     * class C[T] { class D { } ... new D() ... }
     * C[int] c;
     *
     * Then we want to know the class whose where object must
     * be created by C[int]'s adapter class for the
     * AdapterBinding <C[T].D, C>.  That would be C[int].D.
     * That is, we need to take the first element from the binding
     * and instantiate it as <code>inst</code> is.
     */
    public ClassType getInstantiatedClass(ClassType inst,
        AdapterBinding adapter)
    {
        PolyJTypeWrapper tw = null;
        if (inst instanceof SubstType) {
            tw = ((SubstType)inst).substitutor();
        } else {
            throw new InternalCompilerError("Instead of a SubstType, got a "
                + inst.getClass().getName());
        }
        return tw.wrap(adapter.instantiatedType(),
            ts.uninstantiate(inst), inst).toClass();
    }

    public void debug_dumpInstantiations() {
        for (Iterator i = instantiations.iterator(); i.hasNext(); ) {
            ClassType c = (ClassType) i.next();
            System.err.println(c);
        }
    }

    public void debug_dumpNewArrays() {
        for (Iterator i1 = newArrayMethods.entrySet().iterator();
             i1.hasNext();) {
            Map.Entry e= (Map.Entry)i1.next();
            System.err.println(e.getKey()+"\n=============");
            for (Iterator i2 = ((Set)e.getValue()).iterator();
                 i2.hasNext();) {
                int i = ((Integer)i2.next()).intValue();
                System.err.println("\t" + i);
            }
        }
    }
    
    public void debug_dumpAdapterBindings() {
        for (Iterator i1 = adapterBindings.entrySet().iterator();
             i1.hasNext();) {
            Map.Entry e= (Map.Entry)i1.next();
            System.err.println(e.getKey()+"\n=============");
            for (Iterator i2 = ((Map)e.getValue()).entrySet().iterator();
                 i2.hasNext();) {
                Map.Entry f = (Map.Entry)i2.next();
                System.err.println("\t"+
                                   f.getKey()+
                                   " -> "+f.getValue());
            }
        }
    }
}
